# -*- coding: utf-8 -*-

# Copyright 2009-2012 Richard Dymond (rjdymond@gmail.com)
#
# This file is part of SkoolKit.
#
# SkoolKit is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# SkoolKit is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# SkoolKit. If not, see <http://www.gnu.org/licenses/>.

from . import warn, write_line, wrap, parse_int
from .skoolparser import get_address, TABLE_MARKER, TABLE_END_MARKER, LIST_MARKER, LIST_END_MARKER
from .skoolasm import UDGTABLE_MARKER
from .skoolctl import AD_START, AD_WRITER, AD_ORG, AD_END, AD_SET
from .disassembler import Disassembler, HEX4FMT
from .ctlparser import CtlParser

OP_WIDTH = 13

def generate_ctls(snapshot, start):
    ctls = {}
    ctls[start] = 'c'

    # Look for potential 'RET', 'JR d' and 'JP nn' instructions and assume that
    # they end a block (after which another block follows); note that we don't
    # bother examining the byte at 65535 because no block can follow it
    for address in range(start, 65535):
        b = snapshot[address]
        if b == 201:
            ctls[address + 1] = 'c'
        elif b == 195 and address < 65533:
            ctls[address + 3] = 'c'
        elif b == 24 and address < 65534:
            ctls[address + 2] = 'c'

    ctl_parser = CtlParser()
    ctl_parser.ctls = ctls
    disassembly = Disassembly(snapshot, ctl_parser)

    # Scan the disassembly for pairs of adjacent blocks that overlap, and join
    # such pairs
    while True:
        done = True
        for entry in disassembly.entries[:-1]:
            if entry.bad_blocks:
                del ctls[entry.next.address]
                done = False
        if done:
            break
        disassembly.build()

    # Scan the disassembly for blocks that don't end in a 'RET', 'JP nn' or
    # 'JR d' instruction, and join them to the next block
    changed = False
    for entry in disassembly.entries[:-1]:
        last_instr = entry.instructions[-1].operation
        if last_instr != 'RET' and not (last_instr[:2] in ('JP', 'JR') and last_instr[3:].isdigit()):
            del ctls[entry.next.address]
            changed = True
    if changed:
        disassembly.build()

    # Scan the disassembly for pairs of adjacent blocks where the start address
    # of the second block is JRed or JPed to from the first block, and join
    # such pairs
    while True:
        done = True
        for entry in disassembly.entries[:-1]:
            for instruction in entry.instructions:
                operation = instruction.operation
                if operation[:2] in ('JR', 'JP') and operation[-5:] == str(entry.next.address):
                    del ctls[entry.next.address]
                    done = False
                    break
        if done:
            break
        disassembly.build()

    # Mark any NOP sequences at the beginning of a block as a separate zero
    # block
    for entry in disassembly.entries:
        if entry.instructions[0].operation != 'NOP':
            continue
        for instruction in entry.instructions[1:]:
            if instruction.operation != 'NOP':
                break
        ctls[entry.address] = 'z'
        if entry.instructions[-1].operation != 'NOP':
            ctls[instruction.address] = 'c'

    return ctls

def write_ctl(ctlfile, ctls, ctl_hex):
    # Write a control file
    addr_fmt = HEX4FMT if ctl_hex else '{0}'
    with open(ctlfile, 'w') as f:
        for address in sorted(ctls.keys()):
            f.write('{0} {1}\n'.format(ctls[address], addr_fmt.format(address)))

def analyse_blocks(snapshot, ctls):
    # Determine the block start and end addresses
    blocks = [[ctls[address], address, None] for address in sorted(ctls.keys())]
    for i, block in enumerate(blocks[1:]):
        blocks[i][2] = block[1]
    blocks[-1][2] = 65536

    # See which blocks marked as code look like text
    chars = ' ,.abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ'
    for ctl, start, end in blocks:
        if ctl == 'c' and end - start > 5:
            letters = []
            punc = []
            for address in range(start, end):
                char = chr(snapshot[address])
                if char in chars:
                    if char in ',.':
                        punc.append(char)
                    else:
                        letters.append(char)
                else:
                    length = len(punc) + len(letters)
                    if length > 5 and len(set(letters)) >= length / 2 and len(punc) <= length / 5:
                        ctls[start] = 'T'
                        break
                    letters[:] = []
                    punc[:] = []

    # See which blocks marked as code look like data
    for ctl, start, end in blocks:
        if ctl == 'c' and end - start > 3:
            sequences = []
            sequence = [snapshot[start]]
            for b in snapshot[start + 1:end]:
                if b == sequence[0]:
                    sequence.append(b)
                else:
                    sequences.append((sequence[0], len(sequence)))
                    sequence = [b]
            for byte, length in sequences:
                if length > 3:
                    ctls[start] = 'B'
                    break

class Entry:
    def __init__(self, title, ctl, blocks, registers, end_comment, asm_directives):
        self.title = title
        self.ctl = ctl
        self.blocks = blocks
        self.instructions = []
        for block in blocks:
            for instruction in block.instructions:
                instruction.entry = self
                self.instructions.append(instruction)
        first_instruction = self.instructions[0]
        first_instruction.ctl = ctl
        self.registers = registers
        self.end_comment = end_comment
        self.start = None
        self.writer = None
        self.org = None
        self.end = None
        self.properties = []
        for directive, value in asm_directives:
            if directive == AD_START:
                self.start = True
            elif directive == AD_WRITER:
                self.writer = value
            elif directive == AD_ORG:
                self.org = value
            elif directive == AD_END:
                self.end = True
            elif directive.startswith(AD_SET):
                self.properties.append((directive[len(AD_SET):], value))
        self.address = first_instruction.address
        self.description = blocks[0].header
        blocks[0].header = None
        self.next = None
        self.bad_blocks = []
        for block in self.blocks:
            last_instruction = block.instructions[-1]
            if last_instruction.address + last_instruction.size() > block.end:
                self.bad_blocks.append(block)

    def width(self):
        return max([len(i.operation) for i in self.instructions])

class Disassembly:
    def __init__(self, snapshot, ctl_parser, final=False, defb_size=8, defb_mod=1, zfill=False, defm_width=66, asm_hex=False, asm_lower=False):
        self.disassembler = Disassembler(snapshot, defb_size, defb_mod, zfill, defm_width, asm_hex, asm_lower)
        self.ctl_parser = ctl_parser
        if asm_hex:
            self.address_fmt = '{0:04x}' if asm_lower else '{0:04X}'
        else:
            self.address_fmt = '{0}'
        self.build(final)

    def build(self, final=False):
        self.instructions = {}
        self.entries = []
        self._create_entries()
        self.org = self.entries[0].address
        if final:
            self._calculate_references()

    def _create_entries(self):
        for block in self.ctl_parser.get_blocks():
            title = self.ctl_parser.get_block_title(block.start)
            if block.ctl in 'cBT':
                if not title:
                    alt = ''
                    if block.ctl == 'B':
                        alt = '/data?'
                    elif block.ctl == 'T':
                        alt = '/text?'
                    title = 'Routine{0} at {1}'.format(alt, self.address_str(block.start))
                    block.ctl = 'c'
            elif block.ctl in 'bw':
                title = title or 'Data block at {0}'.format(self.address_str(block.start))
            elif block.ctl == 't':
                title = title or 'Message at {0}'.format(self.address_str(block.start))
            elif block.ctl == 'g':
                title = title or 'Game status buffer entry at {0}'.format(self.address_str(block.start))
            elif block.ctl in 'uz':
                title = title or 'Unused'
            for sub_block in block.blocks:
                if sub_block.ctl in 'cBT':
                    instructions = self.disassembler.disassemble(sub_block.start, sub_block.end)
                elif sub_block.ctl in 'bgtuwz':
                    address = sub_block.start
                    lengths = self.ctl_parser.get_lengths(address)
                    one_line = True
                    if not lengths:
                        lengths = [(sub_block.end - sub_block.start, None)]
                        one_line = False
                    instructions = []
                    while address < sub_block.end:
                        if lengths:
                            length, sublengths = lengths.pop(0)
                        end = min((address + length, sub_block.end))
                        if sub_block.ctl == 't':
                            instructions += self.disassembler.defm_range(address, end, one_line, sublengths)
                        elif sub_block.ctl == 'w':
                            instructions += self.disassembler.defw_range(address, end, one_line)
                        elif sub_block.ctl == 'z':
                            instructions += self.disassembler.defs(address, end)
                        else:
                            instructions += self.disassembler.defb_range(address, end, one_line, sublengths)
                        address += length
                else:
                    instructions = self.disassembler.ignore(sub_block.start, sub_block.end)
                sub_block.header = self.ctl_parser.get_block_comment(sub_block.start)
                sub_block.comment = self.ctl_parser.get_instruction_comment(sub_block.start)
                sub_block.instructions = instructions
                for instruction in instructions:
                    self.instructions[instruction.address] = instruction
                    instruction.asm_directives = self.ctl_parser.get_instruction_asm_directives(instruction.address)

            sub_blocks = []
            i = 0
            while i < len(block.blocks):
                sub_block = block.blocks[i]
                i += 1
                sub_blocks.append(sub_block)
                end, comment = self.ctl_parser.get_multiline_comment(sub_block.start)
                if comment is not None:
                    sub_block.comment = comment
                    while i < len(block.blocks) and (end is None or block.blocks[i].start <= end):
                        sub_block.instructions += block.blocks[i].instructions
                        sub_block.end = block.end
                        i += 1

            registers = self.ctl_parser.get_registers(block.start)
            end_comment = self.ctl_parser.get_end_comment(block.start)
            asm_directives = self.ctl_parser.get_entry_asm_directives(block.start)
            self.entries.append(Entry(title, block.ctl, sub_blocks, registers, end_comment, asm_directives))
        for i, entry in enumerate(self.entries[1:]):
            self.entries[i].next = entry

    def _calculate_references(self):
        for entry in self.entries:
            if entry.ctl == 'c':
                for instruction in entry.instructions:
                    operation = instruction.operation
                    if operation[:2].upper() in ('DJ', 'JR', 'JP', 'CA'):
                        addr_str = get_address(operation)
                        if addr_str:
                            callee = self.instructions.get(parse_int(addr_str))
                            if callee:
                                callee.add_referrer(entry)

    def address_str(self, address):
        return self.address_fmt.format(address)

class SkoolWriter:
    def __init__(self, snapshot, ctl_parser, defb_size, defb_mod, zfill, defm_width, asm_hex, asm_lower):
        self.disassembly = Disassembly(snapshot, ctl_parser, True, defb_size, defb_mod, zfill, defm_width, asm_hex, asm_lower)
        if asm_hex:
            self.address_fmt = HEX4FMT.lower() if asm_lower else HEX4FMT
        else:
            self.address_fmt = '{0}'
        self.has_start = ctl_parser.contains_entry_asm_directive(AD_START)
        self.has_org = ctl_parser.contains_entry_asm_directive(AD_ORG)

    def address_str(self, address):
        return self.address_fmt.format(address)

    def write_skool(self, write_refs, text):
        if not self.has_start:
            self.write_asm_directive(AD_START)
            if not self.has_org:
                self.write_asm_directive(AD_ORG, self.address_str(self.disassembly.org))
        for entry_index, entry in enumerate(self.disassembly.entries):
            if entry_index > 0:
                write_line('')
            if entry.start:
                self.write_asm_directive(AD_START)
            if entry.writer:
                self.write_asm_directive(AD_WRITER, entry.writer)
            for name, value in entry.properties:
                self.write_asm_directive('{0}{1}'.format(AD_SET, name), value)
            if entry.org is not None:
                self.write_asm_directive(AD_ORG, entry.org)
            if entry.ctl == 'i':
                end = entry.blocks[0].end
                if end < 65536:
                    if entry.title:
                        self.write_comment(entry.title)
                    write_line('{0}{1}'.format(entry.ctl, self.address_str(entry.blocks[0].start)))
                continue # pragma: no cover
            for block in entry.bad_blocks:
                warn('Code block at {0} overlaps the following block at {1}'.format(self.address_str(block.start), self.address_str(block.end)))
            if entry.title:
                self.write_comment(entry.title)
                wrote_desc = False
                if entry.description:
                    self.write_comment('')
                    self.write_paragraphs(entry.description)
                    wrote_desc = True
                elif entry.ctl == 'c':
                    referrers = entry.instructions[0].referrers
                    if referrers and write_refs:
                        self.write_comment('')
                        self.write_referrers('Used by the ', referrers)
                        wrote_desc = True
                if entry.registers:
                    if not wrote_desc:
                        self.write_comment('')
                        self.write_comment('-')
                    self.write_comment('')
                    for reg, desc in entry.registers:
                        write_line('; {0} {1}'.format(reg, desc))
            op_width = max((OP_WIDTH, entry.width()))
            line_width = op_width + 8
            for block in entry.blocks:
                self.write_paragraphs(block.header)
                rowspan = len(block.instructions)
                multi_line = rowspan > 1 and block.comment
                if multi_line:
                    block.comment = '{{{0}}}'.format(block.comment)
                comment_lines = wrap(block.comment, max((77 - line_width, 10)))
                if multi_line and len(comment_lines) < rowspan:
                    comment_lines[-1] = comment_lines[-1][:-1]
                    comment_lines.extend([''] * (rowspan - len(comment_lines) - 1))
                    comment_lines.append('}')
                for i, instruction in enumerate(block.instructions):
                    ctl = instruction.ctl or ' '
                    address = instruction.address
                    operation = instruction.operation
                    if block.comment:
                        comment = comment_lines[i] if i < len(comment_lines) else ''
                    else:
                        comment = self.to_ascii(instruction.bytes) if text and entry.ctl != 't' else ''
                    if write_refs and entry.ctl == 'c' and ctl == '*' and (i > 0 or not block.header):
                        self.write_referrers('This entry point is used by the ', instruction.referrers)
                    for directive, value in instruction.asm_directives:
                        self.write_asm_directive(directive, value)
                    if entry.ctl == 'c' or comment:
                        write_line(('{0}{1} {2} ; {3}'.format(ctl, self.address_str(address), operation.ljust(op_width), comment)).rstrip())
                    else:
                        write_line(('{0}{1} {2}'.format(ctl, self.address_str(address), operation)).rstrip())
                indent = ' ' * line_width
                for j in range(i + 1, len(comment_lines)):
                    write_line('{0}; {1}'.format(indent, comment_lines[j]))
            self.write_paragraphs(entry.end_comment)
            if entry.end:
                self.write_asm_directive(AD_END)

    def write_comment(self, text):
        if text:
            for line in self.wrap(text):
                write_line('; {0}'.format(line))
        else:
            write_line(';')

    def write_paragraphs(self, paragraphs):
        if paragraphs:
            for p in paragraphs[:-1]:
                self.write_comment(p)
                self.write_comment('.')
            self.write_comment(paragraphs[-1])

    def write_referrers(self, prefix, referrers):
        if referrers:
            if len(referrers) == 1:
                infix = 'routine at '
            else:
                infix = 'routines at {0} and '.format(', '.join('#R{0}'.format(self.address_str(r.address)) for r in referrers[:-1]))
            suffix = '#R{0}'.format(self.address_str(referrers[-1].address))
            self.write_comment('{0}{1}{2}.'.format(prefix, infix, suffix))

    def write_asm_directive(self, directive, value=None):
        suffix = '' if value is None else '={0}'.format(value)
        write_line('; @{0}{1}'.format(directive, suffix))

    def to_ascii(self, data):
        chars = ['[']
        for b in data:
            chars.append(chr(b) if 32 <= b < 127 else '.')
        chars.append(']')
        return ''.join(chars)

    def wrap(self, text):
        lines = []
        for line in self.parse_blocks(text):
            lines.extend(wrap(line, 77))
        return lines

    def parse_block(self, text, start):
        indexes = []
        index = text.index(' ', start)
        indexes.append(index)
        index += 1

        # Parse the table rows or list items
        while True:
            start = text.find('{ ', index)
            if start < 0:
                break
            end = text.index(' }', start)
            index = end + 2
            indexes.append(index)

        indexes.append(len(text))
        return indexes

    def parse_blocks(self, text):
        markers = ((TABLE_MARKER, TABLE_END_MARKER), (UDGTABLE_MARKER, TABLE_END_MARKER), (LIST_MARKER, LIST_END_MARKER))
        indexes = []

        # Find table/list markers and row/item definitions
        index = 0
        while True:
            starts = [text.find(marker[0], index) for marker in markers]
            for i, start in enumerate(starts):
                if start >= 0:
                    if start > 0:
                        indexes.append(start - 1)
                    marker, end_marker = markers[i]
                    end = text.index(end_marker, start) + len(end_marker)
                    indexes.extend(self.parse_block(text[:end], start + len(marker)))
                    break
            else:
                break
            index = indexes[-1] + 1

        # Insert newlines
        end = len(text)
        if end not in indexes:
            indexes.append(end)
        indexes.sort()
        lines = []
        start = 0
        for end in indexes:
            lines.append(text[start:end])
            start = end + 1
        return lines
